import shutil
from pathlib import Path
from time import sleep

from jinja2 import Template

from pgkit.application.models import Postgres
from pgkit.application.utils import *
from pgkit.application.db import DB


class Replica(Postgres):
    def __init__(self, master, port, delay, use_separate_receivewal_service=False):
        super().__init__(name=master.name,
                         host='localhost',
                         port=port,
                         dbname='postgres',
                         version=master.version,
                         username='postgres',
                         password=None,
                         slot=None)
        self.master = master
        self.delay = delay
        self.db_location = f'/var/lib/postgresql/{self.version}/{self.name}'
        self.config_location = f'/etc/postgresql/{self.version}/{self.name}'
        self.wal_location = f'/var/lib/postgresql/wals/{self.version}/{self.name}'
        self.use_separate_receivewal_service = use_separate_receivewal_service

    def start_backup(self):
        self.stop()
        self.remove_related_directories()
        self.create_cluster()
        self.stop()
        self.remove_db_directory()
        self.stop_wal_receive_service()
        self.drop_replication_slot()
        self.create_replication_slot()
        if self.use_separate_receivewal_service:
            self.setup_wal_receive_service()
        self.master.force_switch_wal()
        sleep(10)  # time to receive first wal segments
        self.create_base_backup()
        self.configure_recovery_file()
        self.start()

    def create_cluster(self):
        execute_sync(f'pg_createcluster {self.version} {self.name} -p {self.port}', check_returncode=True)

    def stop(self):
        execute_sync(f'pg_ctlcluster {self.version} {self.name} stop')

    def start(self):
        execute_sync(f'pg_ctlcluster {self.version} {self.name} start', check_returncode=True)

    def restart(self):
        execute_sync(f'pg_ctlcluster {self.version} {self.name} restart')

    def remove_db_directory(self):
        execute_sync(f'rm -rf {self.db_location}')

    def remove_config_directory(self):
        execute_sync(f'rm -rf {self.config_location}')

    def remove_existing_wal_files(self):
        execute_sync(f'rm -rf {self.wal_location}')

    def create_replication_slot(self):
        self.master.create_replication_slot()

    def drop_replication_slot(self):
        self.master.drop_replication_slot()

    def stop_wal_receive_service(self):
        stop_service(f'receivewal-{self.version}-{self.name}')

    def remove_related_directories(self):
        self.remove_db_directory()
        self.remove_config_directory()
        self.remove_existing_wal_files()

    def _stop_old_wal_receive_service(self):
        try:
            stop_service(f'receivewal-{self.version}-{self.name}')
        except:
            pass

    def setup_wal_receive_service(self):
        self._stop_old_wal_receive_service()

        template_path = Path(__file__).parent / "../templates/wal-receive-service.service"
        template = Template(read_file(template_path))

        print('Create wal folder')
        create_directory(self.wal_location)

        print('Create wal-receive service')
        service_file = template.render(
            name=self.name,
            version=self.version,
            host=self.master.host,
            port=self.master.port,
            username=self.master.username,
            password=self.master.password,
            slot=self.master.slot,
            destination=self.wal_location
        )
        write_file(f'/etc/systemd/system/receivewal-{self.version}-{self.name}.service', service_file)

        print('Run wal-receive service')
        restart_service(f'receivewal-{self.version}-{self.name}', reload_systemctl=True)

    def create_base_backup(self):
        print('Take basebackup')
        command = f'/usr/lib/postgresql/{self.version}/bin/pg_basebackup' \
                  f' -h {self.master.host}' \
                  f' -p {self.master.port}' \
                  f' -D {self.db_location}' \
                  f' -U {self.master.username}' \
                  f' -v --checkpoint=fast --progress'
        if self.version >= 10:
            if self.use_separate_receivewal_service:
                command += ' --wal-method=none'
            else:
                command += ' --wal-method=stream --slot={}'.format(self.master.slot)
        execute_sync(command, env=[('PGPASSWORD', self.master.password)], no_pipe=True, check_returncode=True)

        print('change owner to postgres')
        chown(self.db_location, 'postgres')

    def recovery(self, target_time):
        self.configure_recovery_file(recovery=True, recovery_target_time=target_time)
        self._rename_partial_wal_file()
        self.restart()

    def get_config_parameter_value(self, parameter):
        try:
            print(f'Getting config parameter {parameter} from master')
            param_val = self.master.get_config_parameter_value(parameter)
            DB.update_config(self.name, {parameter: param_val})
        except Exception as e:
            print('Error getting master parameter value:', e)
            print('Using DB value')
            param_val = DB.get_config(self.name)[parameter]
        return param_val

    def configure_recovery_file(self, recovery=False, recovery_target_time=None):
        template_path = Path(__file__).parent / "../templates/{}-recovery.conf".format(self.version)
        template = Template(read_file(template_path))
        latest = recovery_target_time == 'latest'

        max_connections = self.get_config_parameter_value('max_connections')
        max_worker_processes = self.get_config_parameter_value('max_worker_processes')

        print('Create recovery.conf file')
        recovery_config = template.render(
            name=self.name,
            host=self.master.host,
            port=self.master.port,
            version=self.version,
            username=self.master.username,
            password=self.master.password,
            base_destination=self.db_location,
            wal_destination=self.wal_location,
            slot=self.master.slot,
            delay=self.delay,
            dbname='postgres',
            standby_mode='true',
            standby_port=self.port,
            recovery_mode=recovery,
            recovery_target_time=recovery_target_time,
            latest=latest,
            max_connections=max_connections,
            max_worker_processes=max_worker_processes,
            use_separate_receivewal_service=self.use_separate_receivewal_service
        )

        file_location = f'/etc/postgresql/{self.version}/{self.name}/postgresql.conf' if self.version >= 12 \
            else f'{self.db_location}/recovery.conf'

        write_file(file_location, recovery_config)
        chown(file_location, 'postgres')

        if self.version >= 12:
            touch_file(f'{self.db_location}/recovery.signal')
            touch_file(f'{self.db_location}/standby.signal')

    def dump(self, database_name, output_path, compress=False, compression_level=9):
        if compress:
            execute_sync(
                f'runuser -l postgres'
                f' -c \'/usr/lib/postgresql/{self.version}/bin/pg_dump'
                f' --no-owner'
                f' -p {self.port}'
                f' -U postgres'
                f' -d {database_name}'
                f' | gzip -{compression_level} > {output_path}\''
            )
        else:
            execute_sync(
                f'runuser -l postgres'
                f' -c \'/usr/lib/postgresql/{self.version}/bin/pg_dump'
                f' -f {output_path}'
                f' --no-owner'
                f' -p {self.port}'
                f' -U postgres'
                f' -d {database_name}\''
            )

    def dumpall(self, output_path, compress=False, compression_level=9):
        if compress:
            execute_sync(
                f'runuser -l postgres '
                f'-c \'/usr/lib/postgresql/{self.version}/bin/pg_dumpall'
                f' --no-owner'
                f' -p {self.port}'
                f' -U postgres'
                f' | gzip -{compression_level} > {output_path}\''
            )
        else:
            execute_sync(
                f'runuser -l postgres'
                f' -c \'/usr/lib/postgresql/{self.version}/bin/pg_dumpall'
                f' -f {output_path}'
                f' --no-owner'
                f' -p {self.port}'
                f' -U postgres\''
            )

    def shell(self):
        execute_sync(f'runuser -l postgres -c \'psql -p {self.port}\'', no_pipe=True)

    def promote(self):
        execute_sync(
            f'pg_ctlcluster {self.version} {self.name} promote'
        )
        remove_file(f'/var/lib/postgresql/{self.version}/{self.name}/recovery.signal')
        self.restart()

    def _rename_partial_wal_file(self):
        wal_location_contents = os.listdir(self.wal_location)
        for filename in wal_location_contents:
            if filename.endswith('.partial'):
                destination_path = os.path.join(self.wal_location, removesuffix(filename, '.partial'))
                shutil.copy2(
                    os.path.join(self.wal_location, filename),
                    destination_path,
                )
                chown(destination_path, 'postgres')

    def _get_current_delay(self):
        file_location = f'/etc/postgresql/{self.version}/{self.name}/postgresql.conf' if self.version >= 12 \
            else f'{self.db_location}/recovery.conf'

        config = read_file(file_location)
        for line in config.split('\n'):
            if 'recovery_min_apply_delay' in line:
                return int(line.split(' ')[2].split("min")[0]) / 60

    def _get_replication_slot_status(self):
        return self.master.run_cmd('select active from pg_replication_slots')

    def _get_postgres_logs(self, lines):
        return execute_sync(f'tail -n {lines} /var/log/postgresql/postgresql-{self.version}-{self.name}.log')

    def _get_receive_wal_logs(self, lines):
        return execute_sync(f'tail -n {lines} /var/log/postgresql/receivewal-{self.version}-{self.name}.log')

    def print_status(self):
        print(f'# DELAY: {self._get_current_delay()}h')
        print()
        print('# REPLICATION SLOT:')
        print(self._get_replication_slot_status())
        print()
        print('# POSTGRES REPLICA SERVICE:')
        print(get_service_status(f'postgresql@{self.version}-{self.name}'))
        print()
        print('# RECEIVE WAL SERVICE:')
        print(get_service_status(f'receivewal-{self.version}-{self.name}'))
        print()
        print('# POSTGRES LOGS:')
        print(self._get_postgres_logs(10))
        print()
        print('# RECEIVE WAL LOGS:')
        print(self._get_receive_wal_logs(10))



import yaml

import click
from pgkit.application.db import DB
import pgkit.application.pg as PG


@click.command()
@click.argument('name', required=True)
@click.option('--replica', is_flag=True, default=False)
def shell(name, replica):
    config = DB.get_config(name)
    PG.shell(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
        shell_to_replica=replica
    )


@click.command()
@click.argument('name', required=True)
@click.argument('database_name', required=True)
@click.argument('output_path', required=True)
@click.option('--compress', required=False, is_flag=True)
@click.option('--compression-level', required=False, type=click.Choice(list(map(str, range(1, 10)))))
def dump(name, database_name, output_path, compress, compression_level):
    if not compress and compression_level:
        return click.echo('--compress flag should be given when compression level is specified')
    config = DB.get_config(name)
    PG.dump(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
        database_name=database_name,
        output_path=output_path,
        compress=compress,
        compression_level=compression_level
    )


@click.command()
@click.argument('name', required=True)
@click.argument('output_path', required=True)
@click.option('--compress', required=False, is_flag=True)
@click.option('--compression-level', required=False, type=click.Choice(list(map(str, range(1, 10)))))
def dumpall(name, output_path, compress, compression_level):
    if not compress and compression_level:
        return click.echo('--compress flag should be given when compression level is specified')
    config = DB.get_config(name)
    PG.dumpall(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
        output_path=output_path,
        compress=compress,
        compression_level=compression_level
    )


@click.command()
@click.argument('name', required=True)
def stop(name):
    config = DB.get_config(name)
    PG.stop(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
    )


@click.command()
@click.argument('name', required=True)
def start(name):
    config = DB.get_config(name)
    PG.start(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
    )


@click.command()
@click.argument('name', required=True)
def restart(name):
    config = DB.get_config(name)
    PG.restart(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
    )


@click.command()
def list():
    return click.echo(yaml.dump(DB.get_configs_list()))



import shlex
import os
import subprocess
import socket
from contextlib import closing


def get_env(env):
    process_env = os.environ.copy()
    for key, value in env:
        process_env[key] = value
    return process_env


def execute(cmd, env=None):
    if env is None:
        env = []
    print('######', cmd, '######')
    popen = subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, universal_newlines=True, env=get_env(env))
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd)


def execute_sync(cmd, env=None, no_pipe=False, timeout=None, check_returncode=False):
    if env is None:
        env = []
    print('######', cmd, '######')
    if no_pipe:
        child = subprocess.Popen(
            shlex.split(cmd),
            env=get_env(env),
        )
    else:
        child = subprocess.Popen(
            shlex.split(cmd),
            env=get_env(env),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
    result = child.communicate(timeout=timeout)
    result = tuple(map(lambda x: x.decode('utf-8').strip() if x else "", result))
    if check_returncode and child.returncode != 0:
        raise subprocess.CalledProcessError(child.returncode, cmd)
    print('*****', result, '*****')

    return result


def read_file(path):
    with open(path) as file_:
        return file_.read()


def write_file(path, text):
    f = open(path, 'w')
    f.write(text)
    f.close()


def create_directory(path, owner='postgres'):
    execute_sync(f'mkdir -p {path}')
    execute_sync(f'chown -R {owner}:{owner} {path}')


def restart_service(service, reload_systemctl=False):
    if reload_systemctl:
        execute_sync('systemctl daemon-reload')
    execute_sync(f'service {service} restart')


def stop_service(service):
    execute_sync(f'service {service} stop')


def get_service_status(service):
    return execute_sync(f'systemctl status {service} --no-pager')


def chown(path, owner):
    execute_sync(f'chown -R {owner}:{owner} {path}')


def get_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def touch_file(path, owner='postgres'):
    execute_sync(f'touch {path}')
    execute_sync(f'chown -R {owner}:{owner} {path}')


def remove_file(path):
    execute_sync(f'rm -f {path}')


def to_number(number):
    if type(number) == str:
        try:
            return int(number)
        except ValueError:
            return float(number)
    return number


def removesuffix(string: str, suffix: str, /) -> str:
    if string.endswith(suffix):
        return string[:-len(suffix)]
    else:
        return string[:]



from tinydb import TinyDB, Query
from pgkit.application.settings import DB_PATH

DEFAULT_MAX_CONNECTIONS = 100
DEFAULT_MAX_WORKERS = 8


class DBClass:
    def __init__(self, db_path):
        self.db = TinyDB(db_path, create_dirs=True)
        self.config_table = self.db.table('config')

    def add_config(self, name, version, host, port, dbname, slot, username, password, replica_port, use_separate_receivewal_service):
        self.config_table.insert({
            'name': name,
            'version': version,
            'host': host,
            'port': port,
            'dbname': dbname,
            'slot': slot,
            'username': username,
            'password': password,
            'replica_port': replica_port,
            'use_separate_receivewal_service': use_separate_receivewal_service,
            'max_connections': DEFAULT_MAX_CONNECTIONS,
            'max_worker_processes': DEFAULT_MAX_WORKERS,
        })

    def remove_config(self, name):
        self.config_table.remove(Query().name == name)

    def get_config(self, name):
        results = self.config_table.search(Query().name == name)
        if not results:
            raise ValueError(f'No config found by the name {name}')
        return self.config_table.search(Query().name == name)[0]

    def get_configs_list(self):
        return [item['name'] for item in self.config_table.all()]

    def update_config(self, name, fields):
        self.config_table.update(fields, Query().name == name)


DB = DBClass(DB_PATH)



from datetime import datetime, timedelta

import click
from dateutil.parser import parse as parse_date
from pytz import timezone
import pgkit.application.pg as PG
from pgkit.application.db import DB


@click.group()
def pitr():
    pass


@pitr.command()
@click.argument('name', required=True)
@click.argument('delay', required=True, type=int)
def backup(name, delay):
    config = DB.get_config(name)
    PG.backup(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
        config.get('use_separate_receivewal_service', False),
        replica_delay=delay
    )


@pitr.command()
@click.argument('name', required=True)
def status(name):
    config = DB.get_config(name)
    PG.print_status(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
    )


@pitr.command()
@click.argument('name', required=True)
@click.argument('target_time', required=True)
def recover(name, target_time):
    config = DB.get_config(name)
    if target_time != 'latest':
        try:
            parsed_target_date = parse_date(target_time)
            target_time = str(timezone('Asia/Tehran').localize(parsed_target_date))
        except ValueError:
            return click.echo('target_time argument should have a valid datetime format.')
    PG.recovery(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
        time_to_recover=target_time
    )


@pitr.command()
@click.argument('name', required=True)
def promote(name):
    config = DB.get_config(name)
    PG.promote(
        config['name'],
        config['host'],
        config['port'],
        config['version'],
        config['dbname'],
        config['username'],
        config['password'],
        config['slot'],
        config['replica_port'],
    )


@pitr.command()
@click.argument('name', required=True)
@click.argument('delay', required=True, type=float)
def status(name, delay):
    target_time = (datetime.today() - timedelta(hours=4.5)).strftime('%Y-%m-%d %H:%M:%S GMT')
    print(f'TARGET TIME: {target_time}')
    if click.confirm('Do you want to continue?'):
        time_to_recover = datetime.today() - timedelta(hours=delay)
        config = DB.get_config(name)
        PG.print_status(
            config['name'],
            config['host'],
            config['port'],
            config['version'],
            config['dbname'],
            config['username'],
            config['password'],
            config['slot'],
            config['replica_port'],
        )



from pgkit.application.utils import execute_sync, to_number


class Postgres:
    def __init__(self, name, host, port, version, dbname, username, password, slot):
        self.name = name
        self.host = host
        self.port = to_number(port)
        self.version = to_number(version)
        self.dbname = dbname
        self.username = username
        self.password = password
        self.slot = slot

    def run_cmd(self, cmd, without_headers=False, timeout=None, check_returncode=False):
        env = [('PGPASSWORD', self.password)] if self.password else []
        return execute_sync(f'psql -h {self.host} -p {self.port} -U {self.username} -d {self.dbname} -c "{cmd}"'
                            f' {"-t" if without_headers else ""}',
                            env=env, timeout=timeout, check_returncode=check_returncode)

    def force_switch_wal(self):
        pg_command = 'select * from pg_switch_xlog()' if self.version < 10 else 'select * from pg_switch_wal()'
        self.run_cmd(pg_command)

    def shell(self):
        env = [('PGPASSWORD', self.password)] if self.password else []
        execute_sync(f'psql -h {self.host} -p {self.port} -U {self.username} -d postgres',
                     env=env, no_pipe=True)

    def create_replication_slot(self):
        args = {'slot_name': f"'{self.slot}'"}
        if self.version != 9.5:
            args['immediately_reserve'] = 'true'
        pg_command = f'select pg_create_physical_replication_slot({", ".join([f"{k} := {v}" for k, v in args.items()])})'
        self.run_cmd(pg_command, check_returncode=True)

    def drop_replication_slot(self):
        pg_command = f"select pg_drop_replication_slot('{self.slot}')"
        self.run_cmd(pg_command)

    def get_config_parameter_value(self, parameter):
        pg_command = f'SHOW {parameter};'
        result = self.run_cmd(pg_command, without_headers=True, timeout=10)
        if result and result[0]:
            return result[0].replace('\n', '').replace('\t', '')
        raise Exception('Cannot get parameter')



import yaml

import click
from pgkit.application.db import DB
from pgkit.application.utils import get_free_port

@click.group()
def config():
    pass


@config.command()
@click.option('--name', help='Postgres Name', required=True, prompt=True)
@click.option('--version', help='Postgres Version', required=True, prompt=True,
              type=click.Choice(['9.5', '10', '11', '12', '13'], case_sensitive=False))
@click.option('--host', help='Host IP', required=True, prompt=True)
@click.option('--port', help='Port Number', required=True, prompt=True, type=int)
@click.option('--dbname', help='Database Name', required=True, prompt=True)
@click.option('--slot', help='Slot Name', required=True, prompt=True)
@click.option('--username', help='Username', required=True, prompt=True)
@click.option('--password', help='Password', required=True, prompt=True, hide_input=True)
@click.option('--replica-port', help='Replica Port', required=False, prompt=False)
@click.option('--use-separate-receivewal-service', help='Use Separate Receivewal Service', required=False, is_flag=True)
def add(name, version, host, port, dbname, slot, username, password, replica_port, use_separate_receivewal_service):
    if not replica_port:
        replica_port = get_free_port()
    DB.add_config(name, version, host, port, dbname, slot, username, password, replica_port, use_separate_receivewal_service)


@config.command()
@click.argument('name', required=True)
@click.option('--dangerous', help='Dangerous Flag', required=True, is_flag=True)
def remove(name, dangerous):
    if not dangerous:
        return click.echo('Removal requires --dangerous flag.')
    DB.remove_config(name)


@config.command()
@click.argument('name', required=True)
def get(name):
    try:
        return click.echo(yaml.dump(dict(DB.get_config(name))))
    except ValueError as e:
        raise click.BadParameter(message=str(e), param=name, param_hint='NAME')



import click
from pgkit.cli.commands.config import config
from pgkit.cli.commands.pitr import pitr
from pgkit.cli.commands.status import status
from pgkit.cli.commands.shellx import shell, dumpall, dump, stop, restart, start, list


@click.group()
def cli():
    pass


def main():
    cli.add_command(config)
    cli.add_command(pitr)
    cli.add_command(status)
    cli.add_command(list)
    cli.add_command(shell)
    cli.add_command(dump)
    cli.add_command(dumpall)
    cli.add_command(start)
    cli.add_command(stop)
    cli.add_command(restart)
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pgkit',
 'pgkit.application',
 'pgkit.application.models',
 'pgkit.application.templates',
 'pgkit.cli',
 'pgkit.cli.commands']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2==3.0.3',
 'PyYAML>=6.0,<7.0',
 'click>=7.1.2,<8.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytz>=2021.1,<2022.0',
 'tinydb>=4.3.0,<5.0.0']

entry_points = \
{'console_scripts': ['pgkit = pgkit.cli.cli:main']}

setup_kwargs = {
    'name': 'pgkit',
    'version': '2.2.2',
    'description': '',
    'long_description': None,
    'author': 'Sadegh Hayeri',
    'author_email': 'hayerisadegh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pgkit.application.utils import execute_sync
from pgkit.application.models import Postgres


class Master(Postgres):
    def __init__(self, name, host, port, version, dbname, username, password, slot):
        super().__init__(name, host, port, version, dbname, username, password, slot)

    def create_replica(self):
        execute_sync('pg_createcluster {} {}'.format(self.version, self.name))

    def remove_replica_directory(self):
        execute_sync('rm -rf {}'.format(self.replica_db_location))

    def start_replica(self):
        execute_sync('pg_ctlcluster {} {} start'.format(self.version, self.name))

    def stop_replica(self):
        execute_sync('pg_ctlcluster {} {} stop'.format(self.version, self.name))



from pgkit.application.models import Master, Replica


def backup(name, host, port, version, dbname, username,
           password, slot, replica_port, use_separate_receivewal_service, replica_delay):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, replica_delay, use_separate_receivewal_service)

    replica.start_backup()


def print_status(name, host, port, version, dbname, username, password, slot, replica_port):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.print_status()


def shell(name, host, port, version, dbname, username, password, slot, replica_port, shell_to_replica):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    if shell_to_replica:
        replica.shell()
    else:
        master.shell()


def recovery(name, host, port, version, dbname, username, password, slot, replica_port, time_to_recover):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.recovery(time_to_recover)


def dump(name, host, port, version, dbname, username, password, slot, replica_port, output_path, database_name,
         compress=False, compression_level=9):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.dump(database_name, output_path, compress, compression_level)


def dumpall(name, host, port, version, dbname, username, password, slot, replica_port, output_path, compress=False,
            compression_level=9):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.dumpall(output_path, compress, compression_level)


def promote(name, host, port, version, dbname, username, password, slot, replica_port):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.promote()


def stop(name, host, port, version, dbname, username, password, slot, replica_port):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.stop()


def start(name, host, port, version, dbname, username, password, slot, replica_port):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.start()


def restart(name, host, port, version, dbname, username, password, slot, replica_port):
    master = Master(name, host, port, version, dbname, username, password, slot)
    replica = Replica(master, replica_port, None)

    replica.restart()
